from unittest import mock

from gcs_migration.models import CredentialYamlDict, PosixCredential, S3Credential

from gcs_migration.check.credential import (
    CredentialCheck,
    UC01,
    UC02,
    UC03,
    UC04,
    UC05,
    UC06,
    UC07,
)


def test_credential_uc01(credential: CredentialYamlDict) -> None:
    uc01 = UC01()

    assert not uc01.is_ok(None)

    cred = PosixCredential(credential)
    assert uc01.is_ok(cred)

    checker = CredentialCheck()
    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, UC01), res)) != []

    res = checker.check(cred)
    assert list(filter(lambda x: isinstance(x, UC01), res)) == []


def test_credential_uc02(credential: CredentialYamlDict) -> None:
    uc02 = UC02()

    assert uc02.is_ok(None)

    cred = PosixCredential(credential)
    cred.username = None
    assert not uc02.is_ok(cred)

    checker = CredentialCheck()
    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, UC02), res)) == []

    res = checker.check(cred)
    assert list(filter(lambda x: isinstance(x, UC02), res)) != []


def test_credential_uc03(credential: CredentialYamlDict) -> None:
    uc03 = UC03()

    assert uc03.is_ok(None)

    cred = PosixCredential(credential)
    cred.identity_id = None
    assert not uc03.is_ok(cred)

    checker = CredentialCheck()
    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, UC03), res)) == []

    res = checker.check(cred)
    assert list(filter(lambda x: isinstance(x, UC03), res)) != []


def test_credential_uc04(credential: CredentialYamlDict) -> None:
    uc04 = UC04()

    assert uc04.is_ok(None)

    cred = PosixCredential(credential)
    cred.identity_username = None
    assert not uc04.is_ok(cred)

    checker = CredentialCheck()
    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, UC04), res)) == []

    res = checker.check(cred)
    assert list(filter(lambda x: isinstance(x, UC04), res)) != []


def test_credential_uc05(credential: CredentialYamlDict) -> None:
    uc05 = UC05()
    checker = CredentialCheck()
    cred = PosixCredential(credential)
    cred.username = None

    assert uc05.is_ok(None)
    assert uc05.is_ok(cred)

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, UC05), res)) == []

    res = checker.check(cred)
    assert list(filter(lambda x: isinstance(x, UC05), res)) == []

    cred.username = "gooduser"
    with mock.patch("gcs_migration.check.credential.pwd.getpwnam"):
        assert uc05.is_ok(cred)
        res = checker.check(cred)
        assert list(filter(lambda x: isinstance(x, UC05), res)) == []

    cred.username = "baduser"
    with mock.patch(
        "gcs_migration.check.credential.pwd.getpwnam", side_effect=KeyError
    ):
        assert not uc05.is_ok(cred)
        res = checker.check(cred)
        assert list(filter(lambda x: isinstance(x, UC05), res)) != []


def test_credential_uc06(credential: CredentialYamlDict) -> None:
    uc06 = UC06()
    checker = CredentialCheck()
    cred = PosixCredential(credential)

    assert uc06.is_ok(None)
    assert uc06.is_ok(cred)
    res = checker.check(cred)
    assert list(filter(lambda x: isinstance(x, UC06), res)) == []

    s3_cred = S3Credential()

    assert not uc06.is_ok(s3_cred)
    res = checker.check(s3_cred)
    assert list(filter(lambda x: isinstance(x, UC06), res)) != []

    s3_cred.s3_key_id = "key_id"

    assert uc06.is_ok(s3_cred)
    res = checker.check(s3_cred)
    assert list(filter(lambda x: isinstance(x, UC06), res)) == []


def test_credential_uc07(credential: CredentialYamlDict) -> None:
    uc07 = UC07()
    checker = CredentialCheck()
    cred = PosixCredential(credential)

    assert uc07.is_ok(None)
    assert uc07.is_ok(cred)
    res = checker.check(cred)
    assert list(filter(lambda x: isinstance(x, UC07), res)) == []

    s3_cred = S3Credential()

    assert not uc07.is_ok(s3_cred)
    res = checker.check(s3_cred)
    assert list(filter(lambda x: isinstance(x, UC07), res)) != []

    s3_cred.s3_secret_key = "secret_key"

    assert uc07.is_ok(s3_cred)
    res = checker.check(s3_cred)
    assert list(filter(lambda x: isinstance(x, UC07), res)) == []
