from typing import List

import pytest

from gcs_migration.models import MigrationPlan

from gcs_migration.check.migration_plan import (
    MigrationPlanCheck,
    MP01,
    MP02,
    MP03,
    MP04,
    MP06,
)


def test_check_mp01(migration_plan: MigrationPlan) -> None:
    mp01 = MP01()
    checker = MigrationPlanCheck()

    assert not mp01.is_ok(None)
    assert mp01.is_ok(migration_plan)

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP01), res)) != []

    res = checker.check(migration_plan)
    assert list(filter(lambda x: isinstance(x, MP01), res)) == []


def test_check_mp02(migration_plan: MigrationPlan) -> None:
    mp02 = MP02()
    checker = MigrationPlanCheck()

    migration_plan.storage_gateway = None

    assert mp02.is_ok(None)
    assert not mp02.is_ok(migration_plan)

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP02), res)) == []

    res = checker.check(migration_plan)
    assert list(filter(lambda x: isinstance(x, MP02), res)) != []


def test_check_mp03(migration_plan: MigrationPlan) -> None:
    mp03 = MP03()
    checker = MigrationPlanCheck()

    migration_plan.mapped_collection = None

    assert mp03.is_ok(None)
    assert not mp03.is_ok(migration_plan)

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP03), res)) == []

    res = checker.check(migration_plan)
    assert list(filter(lambda x: isinstance(x, MP03), res)) != []


@pytest.mark.parametrize(
    "domains,user_pattern,expect",
    [
        ([], "{.username}@example.org", True),
        (["example.org"], "{.username}@example.org", True),
        (["other.org"], "{.username}@example.org", False),
        (["other.org,example.org"], "{.username}@example.org", False),
    ],
)
def test_check_mp04(
    migration_plan: MigrationPlan, domains: List[str], user_pattern: str, expect: bool
) -> None:
    mp04 = MP04()
    checker = MigrationPlanCheck()

    assert migration_plan.storage_gateway is not None
    migration_plan.storage_gateway.allowed_domains = domains

    assert migration_plan.guest_collections is not None
    for gc in migration_plan.guest_collections:
        assert gc.credential is not None
        gc.credential.identity_username = user_pattern.format(gc.credential)

    assert mp04.is_ok(None) is True
    assert mp04.is_ok(migration_plan) == expect

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP04), res)) == []

    res = checker.check(migration_plan)
    assert (list(filter(lambda x: isinstance(x, MP04), res)) == []) == expect


# MP05: "shared endpoints not supported" Deprecated in 0.15.


@pytest.mark.parametrize("disable_guest_collection", [False, True])
def test_check_mp06(
    migration_plan: MigrationPlan, disable_guest_collection: bool
) -> None:
    mp06 = MP06()
    checker = MigrationPlanCheck()

    assert migration_plan.guest_collections is not None
    assert len(migration_plan.guest_collections) > 0

    if disable_guest_collection:
        migration_plan.guest_collections[0].disable = True

    assert mp06.is_ok(None) is True
    assert mp06.is_ok(migration_plan) is not disable_guest_collection

    res = checker.check(None)
    assert list(filter(lambda x: isinstance(x, MP06), res)) == []

    res = checker.check(migration_plan)
    assert (
        list(filter(lambda x: isinstance(x, MP06), res)) == []
    ) != disable_guest_collection
