from unittest import mock

import pytest
import click
from click.testing import CliRunner

from gcs_migration.commands.update import update
from gcs_migration.models import (
    PosixGuestCollection,
    MigrationPlan,
    MyProxyConfiguration,
    PosixCredential,
)
from gcs_migration.parsers import CalloutDict


@pytest.fixture
def migration_plan_existing_extra_collection(
    migration_plan: MigrationPlan,
) -> MigrationPlan:
    credential = PosixCredential()
    credential.username = "new_user"
    credential.subject = "/C=US/O=Globus/CN=New User Certificate"
    credential.gsi_authz_callout_index = 0
    credential.identity_id = "8307e9fd-1735-4552-a4ff-ba2ef886f176"
    credential.identity_username = "new_user@globus.org"

    new_guest_collection = PosixGuestCollection()
    new_guest_collection.host_path = "/host-path"
    new_guest_collection.default_directory = "/host-path"
    new_guest_collection.force_encryption = True
    new_guest_collection.disable_verify = False
    new_guest_collection.force_encryption = True
    new_guest_collection.owner_id = "829320ee-e339-468d-b285-ec91e8789a5f"
    new_guest_collection.owner_string = "new_user@globus.org"
    new_guest_collection.credential = credential
    new_guest_collection.sharing_uuid = "7d91c532-99b4-4946-8c91-f03248bb5a1e"
    new_guest_collection.enable_https = False
    new_guest_collection.share_state_file = (
        "/home/new_user/.globus/sharing/7d91c532-99b4-4946-8c91-f03248bb5a1e.conf"
    )

    migration_plan.add_guest_collection(new_guest_collection)
    return migration_plan


@pytest.fixture
def migration_plan_disabled_collection(
    migration_plan: MigrationPlan,
) -> MigrationPlan:
    assert migration_plan.guest_collections is not None

    for gc in migration_plan.guest_collections:
        gc.disable = True
        break
    return migration_plan


@pytest.fixture
def migration_plan_existing_fewer_collection(
    migration_plan: MigrationPlan,
) -> MigrationPlan:
    assert migration_plan.guest_collections is not None

    migration_plan.guest_collections.pop(0)
    return migration_plan


@pytest.fixture
def migration_plan_different_environment(
    migration_plan: MigrationPlan,
) -> MigrationPlan:
    migration_plan.environment = "preview"
    return migration_plan


@pytest.fixture
def migration_plan_extra_myproxy(migration_plan: MigrationPlan) -> MigrationPlan:
    extra_myproxy = MyProxyConfiguration()
    extra_myproxy.dn = "/C=US/O=Globus/CN=Extra MyProxy"

    c: CalloutDict = {
        "callout_type": "globus_mapping",
        "mandatory": False,
        "library": "libglobus_gridmap_verify_myproxy_callout",
        "symbol": "globus_gridmap_verify_myproxy_callout",
        "environment": {"GLOBUS_MYPROXY_CA_CERT": "/some/new/cert.pem"},
        "raw": "",
    }
    env_string = " ".join([f'{k}="{v}"' for k, v in c["environment"].items()])
    raw = f"{c['callout_type']} {c['library']} {c['symbol']} ENV:{env_string}"
    c["raw"] = raw

    extra_myproxy.callout = c
    assert migration_plan.gsi_authz_configuration is not None
    assert migration_plan.gsi_authz_configuration.callouts is not None
    migration_plan.gsi_authz_configuration.callouts.append(extra_myproxy)
    return migration_plan


def test_update_add_collection(
    migration_plan_cmd_create: MigrationPlan,
    migration_plan_existing_fewer_collection: MigrationPlan,
) -> None:
    assert migration_plan_existing_fewer_collection.guest_collections is not None
    fewer_num_guests = len(migration_plan_existing_fewer_collection.guest_collections)

    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump"), mock.patch(
        "gcs_migration.commands.update.click", wraps=click
    ) as mock_click:
        mock_migration_load.return_value = migration_plan_existing_fewer_collection
        mock_click.get_current_context.return_value.invoke.return_value = (
            migration_plan_cmd_create
        )
        runner = CliRunner()
        result = runner.invoke(update)
        assert result.exit_code == 0
        assert migration_plan_existing_fewer_collection.guest_collections is not None
        assert migration_plan_cmd_create.guest_collections is not None
        assert len(migration_plan_existing_fewer_collection.guest_collections) == len(
            migration_plan_cmd_create.guest_collections
        )
        assert (
            len(migration_plan_existing_fewer_collection.guest_collections)
            > fewer_num_guests
        )


def test_update_remove_collection(
    migration_plan_cmd_create: MigrationPlan,
    migration_plan_existing_extra_collection: MigrationPlan,
) -> None:
    assert migration_plan_existing_extra_collection.guest_collections is not None
    more_num_guests = len(migration_plan_existing_extra_collection.guest_collections)

    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump"), mock.patch(
        "gcs_migration.commands.update.click", wraps=click
    ) as mock_click:
        mock_migration_load.return_value = migration_plan_existing_extra_collection
        mock_click.get_current_context.return_value.invoke.return_value = (
            migration_plan_cmd_create
        )

        runner = CliRunner()
        result = runner.invoke(update)
        assert result.exit_code == 0
        assert migration_plan_existing_extra_collection.guest_collections is not None
        assert migration_plan_cmd_create.guest_collections is not None
        assert len(migration_plan_existing_extra_collection.guest_collections) == len(
            migration_plan_cmd_create.guest_collections
        )
        assert (
            len(migration_plan_existing_extra_collection.guest_collections)
            < more_num_guests
        )


def test_update_change_environment(
    migration_plan_cmd_create: MigrationPlan,
    migration_plan_different_environment: MigrationPlan,
) -> None:
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump"), mock.patch(
        "gcs_migration.commands.update.click", wraps=click
    ) as mock_click:
        mock_migration_load.return_value = migration_plan_different_environment
        mock_click.get_current_context.return_value.invoke.return_value = (
            migration_plan_cmd_create
        )

        runner = CliRunner()
        result = runner.invoke(update)
        assert result.exit_code != 0
        assert "environment" in result.output


def test_update_change_gsi_authz_conf(
    migration_plan_cmd_create: MigrationPlan,
    migration_plan_extra_myproxy: MigrationPlan,
) -> None:
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump"), mock.patch(
        "gcs_migration.commands.update.click", wraps=click
    ) as mock_click:
        mock_migration_load.return_value = migration_plan_extra_myproxy
        mock_click.get_current_context.return_value.invoke.return_value = (
            migration_plan_cmd_create
        )

        runner = CliRunner()
        result = runner.invoke(update)
        assert result.exit_code != 0
        assert "gsi-authz.conf" in result.output


def test_update_fix_disabled(
    migration_plan_cmd_create: MigrationPlan,
    migration_plan_disabled_collection: MigrationPlan,
) -> None:
    """
    Disabling guest collections is deprecated in 0.14. Updating should set disable
    to False.
    """
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump"), mock.patch(
        "gcs_migration.commands.update.click", wraps=click
    ) as mock_click:
        mock_migration_load.return_value = migration_plan_disabled_collection
        mock_click.get_current_context.return_value.invoke.return_value = (
            migration_plan_cmd_create
        )
        assert migration_plan_disabled_collection.guest_collections is not None
        assert migration_plan_cmd_create.guest_collections is not None

        disabled_ids = [
            gc.id
            for gc in migration_plan_disabled_collection.guest_collections
            if gc.disable
        ]
        assert len(disabled_ids) > 0

        runner = CliRunner()
        result = runner.invoke(update)
        assert result.exit_code == 0

        assert len(migration_plan_cmd_create.guest_collections) == len(
            migration_plan_disabled_collection.guest_collections
        )

        disabled_ids = [
            gc.id
            for gc in migration_plan_disabled_collection.guest_collections
            if gc.disable
        ]
        assert len(disabled_ids) == 0


@pytest.mark.parametrize(
    "inherited_prop,value",
    [
        ("force_encryption", True),
        ("force_encryption", False),
        ("force_verify", True),
        ("force_verify", False),
    ],
)
def test_update_change_inherited(
    migration_plan_cmd_create: MigrationPlan,
    inherited_prop: str,
    value: bool,
) -> None:
    with mock.patch.object(
        MigrationPlan, "load"
    ) as mock_migration_load, mock.patch.object(MigrationPlan, "dump"), mock.patch(
        "gcs_migration.commands.update.click", wraps=click
    ) as mock_click:
        loaded_migration_plan = migration_plan_cmd_create
        updated_migration_plan = MigrationPlan()
        updated_migration_plan.__setstate__(migration_plan_cmd_create.__getstate__())
        setattr(loaded_migration_plan.mapped_collection, inherited_prop, not value)
        assert loaded_migration_plan.guest_collections
        for gc in loaded_migration_plan.guest_collections:
            setattr(gc, inherited_prop, not value)

        setattr(updated_migration_plan.mapped_collection, inherited_prop, value)
        mock_migration_load.return_value = loaded_migration_plan
        mock_click.get_current_context.return_value.invoke.return_value = (
            updated_migration_plan
        )

        runner = CliRunner()
        result = runner.invoke(update)
        assert result.exit_code == 0

        if value:
            assert (
                getattr(loaded_migration_plan.mapped_collection, inherited_prop)
                == value
            )
            for gc in loaded_migration_plan.guest_collections:
                assert getattr(gc, inherited_prop) == value
