import json
from typing import Any, Callable, Dict, Generator, Iterable
from unittest import mock

import pytest
import responses
from responses import matchers

import globus_sdk
from globus_sdk import AuthClient, RefreshTokenAuthorizer, TransferClient

from gcs_migration.typing.transfer_documents import (
    AccessDocumentDict,
    AccessListDocumentDict,
    AnonymousEffectivePermissionsDict,
    EndpointDocumentDict,
    EndpointServerDocumentDict,
    RoleDocumentDict,
    RoleListDocumentDict,
    SharedEndpointListDict,
)
from gcs_migration.typing.auth_documents import (
    IdentityDocumentDict,
    IdentityProviderDocumentDict,
)


@pytest.fixture(autouse=True)
def mock_exponential_backoff() -> Generator[None, None, None]:
    with mock.patch("globus_sdk.transport.requests.time"):
        yield


@pytest.fixture
def mock_responses() -> Generator[None, None, None]:
    responses.start()
    yield
    responses.stop()
    responses.reset()


@pytest.fixture
def mock_tokens(mock_responses: None) -> Generator[None, None, None]:
    with mock.patch.object(RefreshTokenAuthorizer, "__new__", return_value=None):
        yield


@pytest.fixture
def mock_get_transfer_api(
    mock_tokens: None,
    host_endpoint_uuid: str,
    share_endpoint_uuid: str,
    v5_endpoint_uuid: str,
    v5_collection_uuid: str,
    access_acl_document: AccessDocumentDict,
    endpoint_server_document: EndpointServerDocumentDict,
    host_endpoint_document: EndpointDocumentDict,
    v5_endpoint_document: EndpointDocumentDict,
    v5_collection_document: EndpointDocumentDict,
    shared_endpoint_document: EndpointDocumentDict,
    role_document: RoleDocumentDict,
    role_list_document: RoleListDocumentDict,
    share_role_list_document: RoleListDocumentDict,
    anonymous_effective_permissions_document: AnonymousEffectivePermissionsDict,
    access_list_document: AccessListDocumentDict,
    shared_endpoint_list_document: SharedEndpointListDict,
) -> None:
    tc = TransferClient()

    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{host_endpoint_uuid}",
        json=host_endpoint_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{share_endpoint_uuid}",
        json=shared_endpoint_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{v5_endpoint_uuid}",
        json=v5_endpoint_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{v5_collection_uuid}",
        json=v5_collection_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{host_endpoint_uuid}/role_list",
        json=role_list_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{host_endpoint_uuid}/shared_endpoint_list",
        json=shared_endpoint_list_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint_manager/endpoint/{share_endpoint_uuid}/access_list",
        json=access_list_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}endpoint/{share_endpoint_uuid}/role_list",
        json=share_role_list_document,
    )
    responses.add(
        "GET",
        f"{tc.base_url}private/endpoint/{share_endpoint_uuid}/anonymous_effective_permissions",
        json=anonymous_effective_permissions_document,
    )
    return None


@pytest.fixture
def unknown_identity_id() -> str:
    return "3f2c5a46-2b93-413d-a0ce-95bd13d075a0"


@pytest.fixture(scope="session")
def auth_client() -> AuthClient:
    return AuthClient(environment="production")


@pytest.fixture
def mock_get_identities_response(auth_client: AuthClient) -> Callable:
    def _mock_get_identities_response(id_doc: IdentityDocumentDict) -> None:
        for provision in [None, True, False]:
            query_params: Dict[str, Any] = {"usernames": id_doc["username"]}
            if provision is not None:
                query_params["provision"] = json.dumps(provision)

            responses.add(
                "GET",
                f"{auth_client.base_url}v2/api/identities",
                match=[matchers.query_param_matcher(query_params)],
                json={"identities": [id_doc]},
            )

            query_params = {"ids": id_doc["id"]}
            if provision is not None:
                query_params["provision"] = json.dumps(provision)

            responses.add(
                "GET",
                f"{auth_client.base_url}v2/api/identities",
                match=[matchers.query_param_matcher(query_params)],
                json={"identities": [id_doc]},
            )

    return _mock_get_identities_response


@pytest.fixture
def mock_get_identity_providers_response(auth_client: AuthClient) -> Callable:
    def _mock_get_identity_providers(
        idp_docs: Iterable[IdentityProviderDocumentDict],
    ) -> None:
        responses.add(
            "GET",
            f"{auth_client.base_url}v2/api/identity_providers",
            match=[
                matchers.query_param_matcher(
                    {
                        "domains": ",".join(
                            dom for idp_doc in idp_docs for dom in idp_doc["domains"]
                        )
                    }
                ),
            ],
            json={"identity_providers": idp_docs},
        )

    return _mock_get_identity_providers


@pytest.fixture
def mock_get_auth_api(
    mock_tokens: None,
    globus_identity_provider_document: IdentityProviderDocumentDict,
    email_identity_provider_document: IdentityProviderDocumentDict,
    example_identity_provider_document: IdentityProviderDocumentDict,
    identity_document: IdentityDocumentDict,
    unused_idp_identity_document: IdentityDocumentDict,
    closed_idp_identity_document: IdentityDocumentDict,
    sharing_local_identity_document: IdentityDocumentDict,
    email_idp_identity_document: IdentityDocumentDict,
    unknown_identity_id: str,
    mock_get_identities_response: Callable[[IdentityDocumentDict], None],
    mock_get_identity_providers_response: Callable[
        [Iterable[IdentityProviderDocumentDict]], None
    ],
    auth_client: AuthClient,
) -> Generator:
    # commands/gridmap
    mock_get_identities_response(identity_document)

    # commands/gridmap
    mock_get_identities_response(email_idp_identity_document)

    mock_get_identities_response(sharing_local_identity_document)

    # check/guest_collection
    responses.add(
        "GET",
        f"{auth_client.base_url}v2/api/identities",
        match=[matchers.query_param_matcher({"ids": "not-a-uuid"})],
        json={"identities": []},
    )
    # check/guest_collection
    responses.add(
        "GET",
        f"{auth_client.base_url}v2/api/identities",
        match=[matchers.query_param_matcher({"ids": "unknown_identity_id}"})],
        json={"identities": []},
    )
    # check/guest_collection
    mock_get_identities_response(closed_idp_identity_document)

    # commands/myproxy
    mock_get_identity_providers_response([globus_identity_provider_document])

    # commands/custom
    mock_get_identity_providers_response([globus_identity_provider_document])
    mock_get_identity_providers_response(
        [globus_identity_provider_document, example_identity_provider_document]
    )

    # commands/custom
    mock_get_identity_providers_response(
        [example_identity_provider_document, globus_identity_provider_document]
    )

    # commands/custom
    mock_get_identity_providers_response([email_identity_provider_document])

    # commands/custom
    responses.add(
        "GET",
        f"{auth_client.base_url}v2/api/identity_providers",
        match=[matchers.query_param_matcher({"domains": "unknown.org"})],
        json={"identity_providers": []},
    )

    with mock.patch(
        "gcs_migration.util.globus_client.globus_sdk", wraps=globus_sdk
    ) as mock_sdk, mock.patch(
        "gcs_migration.util.globus_client.issubclass", return_value=True
    ):
        mock_sdk.AuthClient.return_value = auth_client
        yield
