from typing import Any, Callable, List, Optional
from typing_extensions import TypedDict

import pytest

from gcs_migration.parsers import GridFTPOptionParser
from gcs_migration.models import PosixStorageGateway, S3StorageGateway
from gcs_migration.typing.gcs_documents import IdentityMappingDict
from gcs_migration.typing.transfer_documents import EndpointDocumentDict
from .gridftp_config import (
    GridFTPConfigFixture,
    GridFTPConfigFixtureGenerator,
    GridFTPConfigPolicy,
)


@pytest.fixture
def storage_gateway_display_name() -> str:
    return "Migration Storage Gateway"


@pytest.fixture(params=[None, "user1", "user1,user2,user3"])
def storage_gateway_users_allow(request: Any) -> Optional[str]:
    return request.param


@pytest.fixture(params=[None, "user4", "user5,user6,user7"])
def storage_gateway_users_deny(request: Any) -> Optional[str]:
    return request.param


@pytest.fixture(params=[None, "group1", "group1,group2,group3"])
def storage_gateway_groups_allow(request: Any) -> Optional[str]:
    return request.param


@pytest.fixture(params=[None, "group4", "group5,group6,group7"])
def storage_gateway_groups_deny(request: Any) -> Optional[str]:
    return request.param


class UserPoliciesDict(TypedDict):
    users_allow: List[str]
    users_deny: List[str]
    groups_allow: List[str]
    groups_deny: List[str]


@pytest.fixture
def storage_gateway_user_policies() -> UserPoliciesDict:
    return {
        "users_allow": ["user1", "user2", "user3"],
        "users_deny": ["user5", "user6", "user7"],
        "groups_allow": ["group1", "group2", "group3"],
        "groups_deny": ["group5", "group6", "group7"],
    }


@pytest.fixture
def storage_gateway_allowed_domain() -> str:
    return "one.example.edu"


@pytest.fixture(
    params=[
        None,
        ["one.example.edu"],
        ["one.example.edu", "two.example.edu", "three.example.edu"],
    ]
)
def storage_gateway_allowed_domains(request: Any) -> Optional[List[str]]:
    return request.param


@pytest.fixture
def storage_gateway_identity_mapping() -> "List[IdentityMappingDict]":
    return [
        {
            "DATA_TYPE": "expression_identity_mapping#1.0.0",
            "mappings": [
                {
                    "source": "{username}",
                    "match": "(.*)@example\\.edu",
                    "output": "{0}",
                }
            ],
        }
    ]


@pytest.fixture(
    params=[
        None,
        [
            {
                "DATA_TYPE": "expression_identity_mapping#1.0.0",
                "source": "{username}",
                "match": "(.*)@example\\.edu",
                "output": "{0}",
            },
            {
                "DATA_TYPE": "expression_identity_mapping#1.0.0",
                "source": "{username}",
                "match": "(.*)@two.example\\.edu",
                "output": "{0}",
            },
        ],
    ]
)
def storage_gateway_identity_mappings(
    request: Any,
) -> "Optional[List[IdentityMappingDict]]":
    return request.param


@pytest.fixture
def storage_gateway_gridftp_config(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    gridftp_conf: GridFTPConfigPolicy,
    gridftp_config_gcs: GridFTPConfigPolicy,
    gridftp_config_authorization_myproxy: GridFTPConfigPolicy,
    gridftp_config_credential: GridFTPConfigPolicy,
    gridftp_config_logging: GridFTPConfigPolicy,
    gridftp_config_sharing_policies: GridFTPConfigPolicy,
    gridftp_config_restrict_paths: GridFTPConfigPolicy,
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], GridFTPOptionParser
    ],
) -> GridFTPOptionParser:
    configs = [
        gridftp_config_gcs,
        gridftp_config_authorization_myproxy,
        gridftp_config_credential,
        gridftp_config_logging,
        gridftp_config_sharing_policies,
        gridftp_config_restrict_paths,
    ]

    fixture = gridftp_config_d(configs)
    gridftp_config = fake_gridftp_config_parser_creator(fixture)

    return gridftp_config


@pytest.fixture
def s3_storage_gateway_gridftp_config(
    gridftp_config_d: GridFTPConfigFixtureGenerator,
    gridftp_conf: GridFTPConfigPolicy,
    gridftp_config_gcs: GridFTPConfigPolicy,
    gridftp_config_authorization_myproxy: GridFTPConfigPolicy,
    gridftp_config_credential: GridFTPConfigPolicy,
    gridftp_config_logging: GridFTPConfigPolicy,
    gridftp_config_sharing_policies: GridFTPConfigPolicy,
    gridftp_config_s3_dsi: GridFTPConfigPolicy,
    fake_gridftp_config_parser_creator: Callable[
        [GridFTPConfigFixture], GridFTPOptionParser
    ],
) -> GridFTPOptionParser:
    configs = [
        gridftp_config_gcs,
        gridftp_config_authorization_myproxy,
        gridftp_config_credential,
        gridftp_config_logging,
        gridftp_config_sharing_policies,
        gridftp_config_s3_dsi,
    ]

    fixture = gridftp_config_d(configs)
    gridftp_config = fake_gridftp_config_parser_creator(fixture)

    return gridftp_config


@pytest.fixture
def storage_gateway(
    storage_gateway_display_name: str,
    storage_gateway_user_policies: UserPoliciesDict,
    storage_gateway_allowed_domain: str,
    storage_gateway_identity_mapping: List[IdentityMappingDict],
    storage_gateway_gridftp_config: GridFTPOptionParser,
) -> PosixStorageGateway:
    sg = PosixStorageGateway()
    sg.display_name = storage_gateway_display_name
    sg.users_allow = storage_gateway_user_policies["users_allow"]
    sg.users_deny = storage_gateway_user_policies["users_deny"]
    sg.groups_allow = storage_gateway_user_policies["groups_allow"]
    sg.groups_deny = storage_gateway_user_policies["groups_deny"]
    sg.allowed_domains = [storage_gateway_allowed_domain]
    sg.identity_mappings = storage_gateway_identity_mapping
    sg.gridftp_config = storage_gateway_gridftp_config

    return sg


@pytest.fixture
def s3_storage_gateway(
    storage_gateway_display_name: str,
    storage_gateway_user_policies: UserPoliciesDict,
    storage_gateway_allowed_domain: str,
    storage_gateway_identity_mapping: List[IdentityMappingDict],
    s3_storage_gateway_gridftp_config: GridFTPOptionParser,
    s3_map_filename: str,
) -> S3StorageGateway:
    sg = S3StorageGateway()
    sg.display_name = storage_gateway_display_name
    sg.users_allow = storage_gateway_user_policies["users_allow"]
    sg.users_deny = storage_gateway_user_policies["users_deny"]
    sg.s3_user_credential_required = True
    sg.s3_endpoint = "https://s3.amazonaws.com"
    sg.s3_map_filename = s3_map_filename
    sg.allowed_domains = [storage_gateway_allowed_domain]
    sg.identity_mappings = storage_gateway_identity_mapping
    sg.gridftp_config = s3_storage_gateway_gridftp_config

    return sg


@pytest.fixture
def storage_gateway_cmd_create(
    gridftp_config_parser: GridFTPOptionParser,
    host_endpoint_document: EndpointDocumentDict,
) -> PosixStorageGateway:
    sg = PosixStorageGateway()
    sg.gridftp_config = gridftp_config_parser
    sg.display_name = host_endpoint_document["display_name"]

    return sg
